from abc import ABC, abstractmethod
from typing import TypeVar

from aio_pika.abc import AbstractIncomingMessage

T = TypeVar('T')


class BaseTransaction(ABC):
    @abstractmethod
    async def commit(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def rollback(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def is_done(self) -> bool:
        raise NotImplemented


class RabbitMQIncomingMessageTransaction(BaseTransaction):
    def __init__(
        self,
        incoming_message: AbstractIncomingMessage,
    ) -> None:
        self._incoming_message = incoming_message

        self._is_done = False

    async def commit(self) -> None:
        await self._incoming_message.ack()
        self._is_done = True

    async def rollback(self) -> None:
        await self._incoming_message.nack()
        self._is_done = True

    async def is_done(self) -> bool:
        return self._is_done


class EmptyTransaction(BaseTransaction):
    def __init__(self) -> None:
        self._is_done = False

    async def commit(self) -> None:
        self._is_done = True

    async def rollback(self) -> None:
        self._is_done = True

    async def is_done(self) -> bool:
        return self._is_done



from typing import Dict, List

from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseOutputDeviceManager
from .device_manager import RabbitMQDeviceManager
from .output_device import RabbitMQOutputDevice


class RabbitMQOutputDeviceManager(
    RabbitMQDeviceManager,
    RabbitMQBaseOutputDeviceManager,
):
    def __init__(
        self,
        hosts: List[str],
        user: str,
        password: str,
        vhost: str,
        exchange_name: str,
        publisher_confirms: bool = True,
        channel_qos_kwargs: Dict[str, int | float | bool | None] = None,
        use_transaction: bool = False,
        use_ssl: bool = False,
        port: int = DEFAULT_PORT,
    ):
        super().__init__(
            hosts=hosts,
            user=user,
            password=password,
            vhost=vhost,
            publisher_confirms=publisher_confirms,
            channel_qos_kwargs=channel_qos_kwargs,
            use_transaction=use_transaction,
            use_ssl=use_ssl,
            port=port,
        )
        self._exchange_name = exchange_name

    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQOutputDevice:
        return RabbitMQOutputDevice(
            self,
            device_name,
            self._exchange_name,
        )



from typing import Dict, List

from aio_pika.abc import AbstractQueue

from .device_manager import RabbitMQDeviceManager


async def declare_queue(
    device_manager: RabbitMQDeviceManager,
    queue_name: str,
    passive: bool = False,
    arguments: Dict[str, str | int | List[str]] = None,
    durable: bool = True,
) -> AbstractQueue:
    return await (await device_manager.channel).declare_queue(
        name=queue_name,
        durable=durable,
        passive=passive,
        arguments=arguments,
    )



from io import BytesIO
from typing import Dict, Optional, Tuple

from aio_pika.abc import HeadersType

from .base_device import RabbitMQBaseInputDevice
from .multi_connection_device_manager import RabbitMQMultiConnectionDeviceManager
from .transaction import BaseTransaction, EmptyTransaction, RabbitMQIncomingMessageTransaction


class RabbitMQInputBasicGetDevice(RabbitMQBaseInputDevice):
    def __init__(
        self,
        device_manager: RabbitMQMultiConnectionDeviceManager,
        device_name: str,
        use_transaction: bool,
        max_channels: int,
    ):
        self._device_manager = device_manager
        self._device_name = device_name
        self._use_transaction = use_transaction
        self._max_channels = max_channels
        self._channel_number_to_max_delivery_tag: Dict[int, Optional[int]] = {}

    async def _inner_commit_rollback_all_messages(self, *, commit: bool) -> None:
        if not self._channel_number_to_max_delivery_tag:
            raise Exception("No messages to commit/rollback")

        channel_pool = await self._device_manager.channel
        # noinspection PyProtectedMember
        channels = [await channel_pool._get() for _ in range(self._max_channels)]
        for channel in channels:
            max_delivery_tag_for_channel = self._channel_number_to_max_delivery_tag.pop(channel.number, None)
            if max_delivery_tag_for_channel is None:
                continue

            aiormq_channel = await channel.get_underlay_channel()
            func = aiormq_channel.basic_ack if commit else aiormq_channel.basic_nack
            return await func(delivery_tag=max_delivery_tag_for_channel, multiple=True)

        return None

    async def commit_all_messages(self) -> None:
        return await self._inner_commit_rollback_all_messages(commit=True)

    async def rollback_all_messages(self) -> None:
        return await self._inner_commit_rollback_all_messages(commit=False)

    @property
    def has_messages_available(self) -> bool:
        return False

    async def read(
        self,
    ) -> Optional[Tuple[BytesIO, HeadersType, BaseTransaction]]:
        async with (await self._device_manager.channel).acquire() as channel:
            queue = await channel.get_queue(self._device_name)

            incoming_message = await queue.get(no_ack=not self._use_transaction, fail=False)
            if incoming_message is None:
                return None

            transaction = RabbitMQIncomingMessageTransaction(incoming_message) if self._use_transaction \
                else EmptyTransaction()

            self._channel_number_to_max_delivery_tag[channel.number] = max(
                incoming_message.delivery_tag,
                self._channel_number_to_max_delivery_tag.pop(channel.number, -1),  # -1 because delivery tag is positive
            )
            return BytesIO(incoming_message.body), incoming_message.headers, transaction

    async def connect(self) -> None:
        pass

    async def close(self) -> None:
        pass



from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional, Tuple

from aio_pika.abc import DeliveryMode, HeadersType

from .transaction import BaseTransaction


class RabbitMQBaseDevice(ABC):
    @abstractmethod
    async def connect(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def close(self) -> None:
        raise NotImplemented


class RabbitMQBaseInputDevice(RabbitMQBaseDevice, ABC):
    @abstractmethod
    async def read(
        self,
    ) -> Optional[Tuple[BytesIO, HeadersType, BaseTransaction]]:
        raise NotImplemented

    @abstractmethod
    async def commit_all_messages(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def rollback_all_messages(self) -> None:
        raise NotImplemented

    @property
    @abstractmethod
    def has_messages_available(self) -> bool:
        raise NotImplemented


class RabbitMQBaseOutputDevice(RabbitMQBaseDevice, ABC):
    @abstractmethod
    async def send(
        self,
        stream: BytesIO,
        headers: Optional[HeadersType] = None,
        delivery_mode: DeliveryMode = DeliveryMode.PERSISTENT,
    ) -> bool:
        raise NotImplemented



from typing import List, TypeVar

from pamqp.common import Arguments
from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseInputDeviceManager
from .device_manager import RabbitMQDeviceManager
from .input_consume_device import RabbitMQInputConsumeDevice

T = TypeVar('T')


class RabbitMQConsumeInputDeviceManager(
    RabbitMQDeviceManager,
    RabbitMQBaseInputDeviceManager,
):
    def __init__(
        self,
        hosts: List[str],
        user: str,
        password: str,
        vhost: str,
        prefetch_count: int,
        consumer_arguments: Arguments = None,
        use_transaction: bool = True,
        use_ssl: bool = False,
        port: int = DEFAULT_PORT,
    ):
        super().__init__(
            hosts=hosts,
            user=user,
            password=password,
            vhost=vhost,
            publisher_confirms=True,
            channel_qos_kwargs=dict(prefetch_count=prefetch_count),
            use_transaction=use_transaction,
            use_ssl=use_ssl,
            port=port,
        )
        self._consumer_arguments = consumer_arguments

    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQInputConsumeDevice:
        return RabbitMQInputConsumeDevice(
            self,
            device_name,
            self._use_transaction,
            self._consumer_arguments,
        )



from abc import ABC, abstractmethod

from .base_device import RabbitMQBaseDevice, RabbitMQBaseInputDevice, RabbitMQBaseOutputDevice


class RabbitMQBaseDeviceManager(ABC):
    @abstractmethod
    async def connect(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def close(self) -> None:
        raise NotImplemented

    @abstractmethod
    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQBaseDevice:
        raise NotImplemented


class RabbitMQBaseInputDeviceManager(RabbitMQBaseDeviceManager):
    @abstractmethod
    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQBaseInputDevice:
        raise NotImplemented


class RabbitMQBaseOutputDeviceManager(RabbitMQBaseDeviceManager):
    @abstractmethod
    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQBaseOutputDevice:
        raise NotImplemented



from typing import Dict, List

from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseInputDeviceManager
from .input_device import RabbitMQInputBasicGetDevice
from .multi_connection_device_manager import RabbitMQMultiConnectionDeviceManager


class RabbitMQMultiConnectionBasicGetInputDeviceManager(
    RabbitMQMultiConnectionDeviceManager,
    RabbitMQBaseInputDeviceManager,
):
    def __init__(
        self,
        hosts: List[str],
        user: str,
        password: str,
        vhost: str,
        max_connections: int,
        max_channels: int,
        channel_qos_kwargs: Dict[str, int | float | bool | None] = None,
        use_transaction: bool = True,
        use_ssl: bool = False,
        port: int = DEFAULT_PORT,
    ):
        super().__init__(
            hosts=hosts,
            user=user,
            password=password,
            vhost=vhost,
            publisher_confirms=False,
            max_connections=max_connections,
            max_channels=max_channels,
            channel_qos_kwargs=channel_qos_kwargs,
            use_transaction=use_transaction,
            use_ssl=use_ssl,
            port=port,
        )

    async def get_device(
        self,
        device_name: str,
    ) -> RabbitMQInputBasicGetDevice:
        return RabbitMQInputBasicGetDevice(
            self,
            device_name,
            self._use_transaction,
            self._max_channels,
        )



import random
from abc import ABC
from typing import Dict, List, Optional

from aio_pika import connect_robust
from aio_pika.abc import AbstractRobustChannel, AbstractRobustConnection
from aio_pika.connection import make_url
from aio_pika.pool import Pool
from no_exception import NoException
from pamqp.constants import DEFAULT_PORT

from .device_manager import RabbitMQDeviceManager


class RabbitMQMultiConnectionDeviceManager(RabbitMQDeviceManager, ABC):
    def __init__(
        self,
        hosts: List[str],
        user: str,
        password: str,
        vhost: str,
        publisher_confirms: bool,
        max_connections: int,
        max_channels: int,
        channel_qos_kwargs: Dict[str, int | float | bool | None] = None,
        use_transaction: bool = False,
        use_ssl: bool = False,
        port: int = DEFAULT_PORT,
    ):
        super().__init__(
            hosts=hosts,
            user=user,
            password=password,
            vhost=vhost,
            publisher_confirms=publisher_confirms,
            channel_qos_kwargs=channel_qos_kwargs,
            use_transaction=use_transaction,
            use_ssl=use_ssl,
            port=port,
        )
        self._max_connections = max_connections
        self._max_channels = max_channels

        self._connection: Optional[Pool[AbstractRobustConnection]] = None
        self._channel: Optional[Pool[AbstractRobustChannel]] = None

    @property
    async def connection(self) -> Pool[AbstractRobustConnection]:
        return await super().connection

    @property
    async def channel(self) -> Pool[AbstractRobustChannel]:
        return await super().channel

    async def _create_connection(self) -> None:
        self._connection = Pool(
            connect_robust,
            make_url(
                host=random.choice(self._hosts),
                port=self._port,
                login=self._user,
                password=self._password,
                virtualhost=self._vhost,
                ssl=self._use_ssl,
            ),
            max_size=self._max_connections,
        )

    async def _create_channel(self) -> None:
        async def inner_create_channel() -> AbstractRobustChannel:
            async with (await self.connection).acquire() as connection:
                channel = await connection.channel(
                    publisher_confirms=self._publisher_confirms,
                )
                await channel.set_qos(**self._channel_qos_kwargs)
                return channel

        self._channel = Pool(
            inner_create_channel,
            max_size=self._max_channels,
        )

    async def _close_connection(self) -> None:
        with NoException():
            await self._connection.close()

    async def _close_channel(self) -> None:
        with NoException():
            await self._channel.close()



from .base_device import RabbitMQBaseDevice, RabbitMQBaseInputDevice, RabbitMQBaseOutputDevice
from .base_device_manager import RabbitMQBaseDeviceManager, RabbitMQBaseInputDeviceManager, \
    RabbitMQBaseOutputDeviceManager
from .consume_input_device_manager import RabbitMQConsumeInputDeviceManager
from .create_queue import declare_queue
from .device_manager import RabbitMQDeviceManager
from .input_consume_device import RabbitMQInputConsumeDevice
from .input_device import RabbitMQInputBasicGetDevice
from .multi_connection_device_manager import RabbitMQMultiConnectionDeviceManager
from .multi_connection_input_device_manager import RabbitMQMultiConnectionBasicGetInputDeviceManager
from .output_device import RabbitMQOutputDevice
from .output_device_manager import RabbitMQOutputDeviceManager
from .transaction import BaseTransaction, EmptyTransaction, RabbitMQIncomingMessageTransaction



from io import BytesIO
from typing import Optional

from aio_pika import DeliveryMode, Message, RobustExchange
from aio_pika.abc import HeadersType
from pamqp.commands import Basic

from .base_device import RabbitMQBaseOutputDevice
from .device_manager import RabbitMQDeviceManager


class RabbitMQOutputDevice(RabbitMQBaseOutputDevice):
    def __init__(
        self,
        device_manager: RabbitMQDeviceManager,
        device_name: str,
        exchange_name: str,
    ):
        self._device_manager = device_manager
        self._device_name = device_name
        self._exchange_name = exchange_name

        self._exchange: Optional[RobustExchange] = None

    @property
    async def exchange(self) -> Optional[RobustExchange]:
        if self._exchange is None or self._exchange.channel.is_closed:
            self._exchange = await (await self._device_manager.channel).get_exchange(
                self._exchange_name,
                ensure=self._exchange_name != "",
            )
        return self._exchange

    async def send(
        self,
        stream: BytesIO,
        headers: Optional[HeadersType] = None,
        delivery_mode: DeliveryMode = DeliveryMode.PERSISTENT,
    ) -> bool:
        return isinstance(
            await (await self.exchange).publish(
                Message(
                    body=stream.read(),
                    headers=headers,
                    delivery_mode=delivery_mode,
                ),
                routing_key=self._device_name,
            ),
            Basic.Ack,
        )

    async def connect(self) -> None:
        self._exchange = await self.exchange

    async def close(self) -> None:
        pass
